import java.awt.Color
import java.awt.Paint
import java.text.DecimalFormat
import java.text.NumberFormat
import java.text.SimpleDateFormat

import org.jfree.chart.axis.DateAxis
import org.jfree.chart.axis.DateTickUnit
import org.jfree.chart.axis.DateTickUnitType
import org.jfree.chart.labels.CategoryItemLabelGenerator
import org.jfree.chart.labels.ItemLabelAnchor
import org.jfree.chart.labels.ItemLabelPosition
import org.jfree.chart.labels.StandardCategoryItemLabelGenerator
import org.jfree.chart.labels.StandardPieSectionLabelGenerator
import org.jfree.chart.renderer.category.BarRenderer
import org.jfree.chart.renderer.category.GanttRenderer
import org.jfree.chart.renderer.category.StandardBarPainter
import org.jfree.data.category.CategoryDataset
import org.jfree.data.category.DefaultCategoryDataset
import org.jfree.data.gantt.Task
import org.jfree.data.gantt.TaskSeries
import org.jfree.data.gantt.TaskSeriesCollection
import org.jfree.data.general.DefaultPieDataset
import org.jfree.data.general.PieDataset
import org.jfree.data.time.DateRange
import org.jfree.ui.TextAnchor
import org.joda.time.DateTime
import org.joda.time.DateTimeConstants
import org.joda.time.Hours

import groovy.time.TimeCategory

reportTool = TimeByXReportToolFactory.createTimeByXReportTool(componentAccessor, i18n, tickUnitCalculator, userDateTimeFormatter, dateTimeStyle, user, fieldId, prettyPrint)
filename = "Time by ${reportTool.fieldName} Report.pdf".toString() // "filename" is a standard Java String, not a GString

/**
 * Factor to create Velocity tools that calculate metrics for "Time by X" reports.
 * @see time-by-x-report.vm as the actual template.
 */
class TimeByXReportToolFactory {
	private static final FIELD_STATUS = "status"
	private static final FIELD_ASSIGNEE = "assignee"

	static TimeByXReportTool createTimeByXReportTool(componentAccessor, i18n, tickUnitCalculator, userDateTimeFormatter, dateTimeStyle, user, fieldId, prettyPrint) {
		def reportTool
		switch (fieldId) {
			case FIELD_STATUS:
				reportTool = new TimeByStatusReportTool(
						componentAccessor: componentAccessor,
						i18n: i18n,
						tickUnitCalculator: tickUnitCalculator,
						userDateTimeFormatter: userDateTimeFormatter,
						dateTimeStyle: dateTimeStyle,
						user: user,
						prettyPrint: prettyPrint)
				break

			case FIELD_ASSIGNEE:
				reportTool = new TimeByAssigneeReportTool(
						componentAccessor: componentAccessor,
						i18n: i18n,
						tickUnitCalculator: tickUnitCalculator,
						userDateTimeFormatter: userDateTimeFormatter,
						dateTimeStyle: dateTimeStyle,
						user: user,
						prettyPrint: prettyPrint)
				break

			default:
				reportTool = new TimeByCustomFieldReportTool(
						componentAccessor: componentAccessor,
						i18n: i18n,
						tickUnitCalculator: tickUnitCalculator,
						userDateTimeFormatter: userDateTimeFormatter,
						dateTimeStyle: dateTimeStyle,
						user: user,
						prettyPrint: prettyPrint,
						fieldId: fieldId)
				break
		}

		reportTool.init()

		return reportTool
	}
}

class TimeByStatusReportTool extends TimeByXReportTool {
	private enum StatusColor {
		UNDEFINED("#8993A4", "#42526E"),
		TO_DO("#B3BAC5", "#505F79"),
		IN_PROGRESS("#B3D4FF", "#0052CC"),
		DONE("#79F2C0", "#00875A");

		String bgColor
		String fgColor

		StatusColor(String bgColor, String fgColor) {
			this.bgColor = bgColor
			this.fgColor = fgColor
		}
	}

	private static final FIELD_ID = "status"
	private static final colorsByStatusCategory = ["undefined": StatusColor.UNDEFINED, "new": StatusColor.TO_DO, "indeterminate": StatusColor.IN_PROGRESS, "done": StatusColor.DONE]
	private colorsByStatus = [:]

	def init() {
		initColors()
	}

	private initColors() {
		for (status in componentAccessor.constantsManager.statuses) {
			colorsByStatus[status.name] = colorsByStatusCategory[status.statusCategory.key]
		}
	}

	def getFieldName() {
		i18n.getText("issue.field.status")
	}

	protected def getFieldId() {
		FIELD_ID
	}

	protected def getField() {
		FIELD_ID
	}

	protected def getCurrentFieldValue(def issue) {
		issue.status.name
	}

	def getBgColor(value) {
		getStatusColor(value).bgColor
	}

	def getFgColor(value) {
		getStatusColor(value).fgColor
	}

	private getStatusColor(value) {
		if (colorsByStatus[value]) {
			return colorsByStatus[value]
		}
		def status = componentAccessor.constantsManager.getStatus(value) // fallback mode in case of different locale
		colorsByStatus[value] = status ? colorsByStatusCategory[status.statusCategory.key] : StatusColor.UNDEFINED
		return colorsByStatus[value]
	}

	protected fillPieChartColors(chart) {
		for (key in chart.plot.dataset.keys) {
			chart.plot.setSectionPaint(key, Color.decode(getBgColor(key)))
		}
	}

	protected getBarRenderer(chart) {
		new ColoredByStatusCategoryBarRenderer(chart.plot.dataset.columnKeys)
	}

	protected getGanttChartRenderer(chart) {
		new ColoredByStatusGanttRenderer(chart: chart)
	}

	private class ColoredByStatusCategoryBarRenderer extends BarRenderer {
		def columnKeys

		ColoredByStatusCategoryBarRenderer(columnKeys) {
			this.columnKeys = columnKeys
		}

		Paint getItemPaint(final int row, final int column) {
			Color.decode(getBgColor(columnKeys[column]))
		}
	}

	private class ColoredByStatusGanttRenderer extends GanttRenderer {
		def chart

		Paint getItemPaint(int row, int col) {
			Color.decode(getBgColor(chart.plot.dataset.keys[col].replaceFirst(String.valueOf(col), "")))
		}
	}
}

class TimeByAssigneeReportTool extends TimeByXReportTool {
	protected static final FIELD_ID = "assignee"

	def getFieldName() {
		i18n.getText("issue.field.assignee")
	}

	protected def getFieldId() {
		FIELD_ID
	}

	protected def getField() {
		FIELD_ID
	}

	protected def getCurrentFieldValue(def issue) {
		(issue.assignee && issue.assignee.displayName) ? issue.assignee.displayName : normalizeValue(null)
	}

	protected def normalizeValue(def value) {
		value ?: i18n.getText("common.status.unassigned")
	}

}

class TimeByCustomFieldReportTool extends TimeByXReportTool {
	private static final DecimalFormat floatFormat = new DecimalFormat("#,###.###")

	def fieldId
	private def fieldName
	private def customFieldType

	def getFieldName() {
		fieldName
	}

	protected init() {
		def customField = componentAccessor.customFieldManager.getCustomFieldObject(fieldId)
		customFieldType = customField.customFieldType
		fieldName = customField.name
	}

	protected def getFieldId() {
		fieldId
	}

	protected def getField() {
		fieldName
	}

	protected def getCurrentFieldValue(def issue) {
		def value = issue.getCustomFieldValue(fieldId)

		if (!value) {
			return normalizeValue(value)
		}

		switch (customFieldType.key) {
			case "com.atlassian.jira.plugin.system.customfieldtypes:multicheckboxes":
			case "com.atlassian.jira.plugin.system.customfieldtypes:multiselect":
				return value.collect { it.value }.join(",")

			case "com.atlassian.jira.plugin.system.customfieldtypes:datepicker":
				return userDateTimeFormatter.withDefaultZone().withStyle(dateTimeStyle.DATE).format(value)

			case "com.atlassian.jira.plugin.system.customfieldtypes:datetime":
				return userDateTimeFormatter.withStyle(dateTimeStyle.COMPLETE).format(value)

			case "com.atlassian.jira.plugin.system.customfieldtypes:multigrouppicker":
			case "com.atlassian.jira.plugin.system.customfieldtypes:grouppicker":
			case "com.atlassian.jira.plugin.system.customfieldtypes:multiversion":
			case "com.atlassian.jira.plugin.system.customfieldtypes:version":
			case "com.atlassian.servicedesk:sd-customer-organizations":
			case "com.pyxis.greenhopper.jira:gh-sprint":
				return value.collect { it.name }.join(", ")

			case "com.atlassian.jira.plugin.system.customfieldtypes:labels":
				return value.collect { it.label }.join(" ")

			case "com.atlassian.jira.plugin.system.customfieldtypes:float":
				return floatFormat.format(value)

			case "com.atlassian.jira.plugin.system.customfieldtypes:project":
				return value.name

			case "com.atlassian.jira.plugin.system.customfieldtypes:radiobuttons":
			case "com.atlassian.jira.plugin.system.customfieldtypes:select":
				return value.value

			case "com.atlassian.jira.plugin.system.customfieldtypes:cascadingselect":
				return (value != null) ? value.get(null).toString() + ((value.get("1") != null) ? " - " + value.get("1").toString() : "") : null

			case "com.atlassian.jira.plugin.system.customfieldtypes:multiuserpicker":
			case "com.atlassian.jira.plugin.system.customfieldtypes:people":
			case "com.atlassian.servicedesk:sd-request-participants":
				return value.collect { it.displayName }.join(", ")

			case "com.atlassian.jira.plugin.system.customfieldtypes:userpicker":
				return value.displayName

			case "com.atlassian.servicedesk:vp-origin":
				return value.requestType.name

			case "com.pyxis.greenhopper.jira:gh-epic-link":
				return customFieldType.getEpicDisplayName(value)

			case "com.atlassian.servicedesk:sd-request-feedback":
				return value.rate

			default:
				return value.toString()
		}
	}

	protected def normalizeValue(def value) {
		value ?: "Unset"
	}
}

abstract class TimeByXReportTool {
	private static final String DEFAULT_TEXT_COLOR = "#000000"
	private static final int ONE_DAY_IN_HOURS = 24
	private static final int ONE_YEAR_IN_HOURS = 365 * ONE_DAY_IN_HOURS

	protected static final DecimalFormat hourFormat = new DecimalFormat("#.#h")
	protected static final DecimalFormat hourFormatForPrettyPrint = new DecimalFormat("#h")
	protected static final DecimalFormat percentageFormat = new DecimalFormat("#.#%")

	protected componentAccessor
	protected i18n
	protected tickUnitCalculator
	protected user
	protected prettyPrint
	protected userDateTimeFormatter
	protected dateTimeStyle

	abstract def getFieldName()

	protected abstract def getFieldId()

	protected abstract def getField()

	protected abstract def getCurrentFieldValue(def issue)

	protected init() {
	}

	protected def normalizeValue(def value) {
		value ?: ""
	}

	def getFgColor(value) {
		DEFAULT_TEXT_COLOR
	}

	protected fillPieChartColors(chart) {
		chart.plot.drawingSupplier = new GradientDrawingSupplier(chart.plot.dataset.itemCount)
	}

	protected getBarRenderer(chart) {
		new GradientBarRenderer(chart.plot.dataset.columnKeys.size())
	}

	protected getGanttChartRenderer(chart) {
		new ColoredGanttRenderer(chart: chart)
	}

	private isRelatedChangeItem(changeItem) {
		if (changeItem.hasProperty('fieldId')) {
			return fieldId.equals(changeItem.fieldId)
		}
		return field.equals(changeItem.field)
	}

	/**
	 * @return the list of maps (keyed with "name", "value") that represent the key metric items.
	 */
	def getKeyMetricsByIssue(def issue) {
		def metrics = []

		def now = new Date()
		def issueCreated = issue.getCreated()
		def firstAssignedTime
		def fieldChanges = []

		def changeHistories = componentAccessor.changeHistoryManager.getChangeHistories(issue)
		for (changeHistory in changeHistories) {
			for (changeItem in changeHistory.changeItemBeans) {
				if ((firstAssignedTime == null) && (TimeByAssigneeReportTool.FIELD_ID.equals(changeItem.field))) {
					firstAssignedTime = changeItem.fromString ? issueCreated : changeHistory.timePerformed
				}
				if (isRelatedChangeItem(changeItem)) {
					fieldChanges.add([timePerformed: changeHistory.timePerformed, fromString: changeItem.fromString, toString: changeItem.toString])
				}
			}
		}

		if (firstAssignedTime) {
			firstAssignedTime = getDuration(issueCreated, firstAssignedTime)
			firstAssignedTime = firstAssignedTime ?: "0 second" // assigned upon issue creation
		} else if (issue.assignee) {
			firstAssignedTime = "0 second"
		}

		def lastChanged = !fieldChanges.empty ? fieldChanges.get(fieldChanges.size() - 1) : null
		def lastChangedTo = lastChanged ? normalizeValue(lastChanged.toString) : getCurrentFieldValue(issue)
		def lastChangedTime = lastChanged ? lastChanged.timePerformed : issueCreated
		metrics.add([name: "$fieldName last changed at", "value": userDateTimeFormatter.withStyle(dateTimeStyle.COMPLETE).format(lastChangedTime)])

		def previousChanged = (fieldChanges.size() > 1) ? fieldChanges.get(fieldChanges.size() - 2) : null
		def previousChangedTo = lastChanged ? " (" + normalizeValue(lastChanged.fromString) + ")" : ""
		def previousChangedTime = previousChanged ? previousChanged.timePerformed : (lastChanged ? issueCreated : null)
		metrics.add([name: "Time in previous $fieldName$previousChangedTo", "value": previousChangedTime ? getDuration(previousChangedTime, lastChangedTime) : "No previous value"])

		metrics.add([name: "Time in current $fieldName ($lastChangedTo)", "value": getDuration(lastChangedTime, now)])

		metrics.add([name: "Time to first assignee", "value": firstAssignedTime ?: "Not assigned yet"])

		def firstResponseTime = getIssueFirstResponseTime(issue)
		metrics.add([name: "Time to first comment", value: firstResponseTime ? getDuration(issueCreated, firstResponseTime) : "Not commented yet"])

		def resolutionTime = getIssueResolutionTime(issue)
		metrics.add([name: "Time to resolution", value: resolutionTime ? getDuration(issueCreated, resolutionTime) : "Not resolved yet"])

		return metrics
	}

	/**
	 * @return the list of maps (keyed with "fromString", "toString", "duration", "author", "timePerformed") that represent the change history
	 */
	def getChangeHistoryByIssue(def issue) {
		def start = issue.getCreated()
		def fieldChanges = []

		def changeHistories = componentAccessor.changeHistoryManager.getChangeHistories(issue)
		for (changeHistory in changeHistories) {
			for (changeItem in changeHistory.changeItemBeans) {
				if (isRelatedChangeItem(changeItem)) {
					fieldChanges.add([fromString: normalizeValue(changeItem.fromString), toString: normalizeValue(changeItem.toString), duration: getDuration(start, changeHistory.timePerformed), author: changeHistory.authorDisplayName, timePerformed: changeHistory.timePerformed])

					start = changeHistory.timePerformed
				}
			}
		}

		return fieldChanges
	}

	/**
	 * @return the map where the key is field value and the value is a list of maps (keyed with "totalTime", "totalTimeFormatted", "count", "percentage", "percentageFormatted") that represent the change history.
	 */
	def getTotalTimeSpentMetricsByIssue(def issue) {
		def durations = getDurations(issue)
		def metrics = new LinkedHashMap()

		def grandTotal = 0
		for (key in durations.keySet()) {
			def sum = 0
			for (duration in durations.get(key)) {
				sum += duration
			}

			grandTotal += sum

			def now = new Date()
			metrics[key] = [:]
			metrics[key]["totalTimeFormatted"] = getDuration(new Date(now.getTime() - sum), now)
			metrics[key]["totalTime"] = sum
			metrics[key]["count"] = durations.get(key).size()
		}

		for (key in metrics.keySet()) {
			metrics[key]["percentage"] = metrics[key]["totalTime"] / grandTotal
			metrics[key]["percentageFormatted"] = percentageFormat.format(metrics[key]["percentage"])
		}

		return metrics
	}

	TaskSeriesCollection getChangeHistoryTaskSeriesCollection(issue, fieldChanges) {
		def dataset = new TaskSeriesCollection()

		def series = new TaskSeries("Change history")
		int ndx = 0
		def previousTime = issue.getCreated()

		for (change in fieldChanges) {
			series.add(new Task(ndx++ + change.fromString, previousTime, change.timePerformed))
			previousTime = change.timePerformed
		}

		def lastChange = fieldChanges.get(fieldChanges.size() - 1)
		series.add(new Task(ndx + lastChange.toString, lastChange.timePerformed, new Date()))

		dataset.add(series)
		return dataset
	}

	def getChangeHistoryGanttChartCustomizer(issue) {
		return { chart ->
			chart.plot.renderer = getGanttChartRenderer(chart)
			chart.plot.renderer.barPainter = new StandardBarPainter()
			chart.plot.renderer.baseItemLabelGenerator = new LabelGenerator()
			chart.plot.renderer.basePositiveItemLabelPosition = new ItemLabelPosition(ItemLabelAnchor.OUTSIDE3, TextAnchor.CENTER_LEFT)
			chart.plot.renderer.itemLabelAnchorOffset = 7.0D
			chart.plot.renderer.baseItemLabelFont = chart.plot.renderer.getBaseItemLabelFont().deriveFont(10f)
			chart.plot.renderer.baseItemLabelsVisible = true
			chart.plot.renderer.shadowVisible = false
			chart.plot.renderer.drawBarOutline = true

			chart.plot.outlinePaint = null
			chart.plot.domainAxis.visible = false
			chart.plot.domainAxis.upperMargin = 0D

			// create range axis (X)
			DateAxis rangeAxis = new DateAxis("")

			Date start = issue.getCreated()
			Date end = new Date()

			rangeAxis.lowerMargin = 0D
			rangeAxis.upperMargin = 0.1D
			rangeAxis.rangeWithMargins = new DateRange(start, end)
			rangeAxis.axisLineVisible = false
			rangeAxis.tickMarkInsideLength = 4.0f

			def durationInSeconds = (end.time - start.time) / 1000
			def tickUnitInSeconds = (int) tickUnitCalculator.calculateTime(durationInSeconds)
			rangeAxis.tickUnit = new DateTickUnit(DateTickUnitType.SECOND, tickUnitInSeconds)
			def rangeAxisTickLabelFormat = getAxisTickLabelFormat(start.time, end.time, !isAtRoundDay(tickUnitInSeconds))
			def tickCount = durationInSeconds / tickUnitInSeconds
			rangeAxis.dateFormatOverride = rangeAxisTickLabelFormat.dateFormat
			rangeAxis.verticalTickLabels = tickCount > 10 && rangeAxisTickLabelFormat.isLong

			chart.plot.setRangeAxis(0, rangeAxis)
		}
	}

	private static class LabelGenerator implements CategoryItemLabelGenerator {
		String generateRowLabel(CategoryDataset dataset, int row) {
			""
		}

		String generateColumnLabel(CategoryDataset dataset, int column) {
			""
		}

		String generateLabel(CategoryDataset dataset, int row, int column) {
			return dataset.keys[column].replaceFirst(String.valueOf(column), "")
		}
	}

	private static class ColoredGanttRenderer extends GanttRenderer {
		def chart

		Paint getItemPaint(int row, int col) {
			def max = chart.plot.dataset.keys.size()
			def key = getTrimmedKey(col)
			for (int i = 0; i < col; i++) {
				if (getTrimmedKey(i).equals(key)) {
					return RenderingUtils.getGradientPaint((i + 1) / max)
				}
			}
			return RenderingUtils.getGradientPaint((col + 1) / max)
		}

		private String getTrimmedKey(ndx) {
			chart.plot.dataset.keys[ndx].replaceFirst(String.valueOf(ndx), "")
		}
	}

	CategoryDataset getTotalTimeSpentCategoryDataset(Map totalTimeSpentMetrics) {
		def dataset = new DefaultCategoryDataset()
		for (metrics in totalTimeSpentMetrics.entrySet()) {
			def hours = metrics.value["totalTime"] / 1000 / 60 / 60
			dataset.setValue(hours, "time", metrics.key)
		}

		return dataset
	}

	def getTotalTimeSpentBarChartCustomizer() {
		return { chart ->
			chart.plot.renderer = getBarRenderer(chart)
			chart.plot.renderer.barPainter = new StandardBarPainter()
			chart.plot.renderer.drawBarOutline = true
			chart.plot.renderer.shadowVisible = false

			chart.plot.renderer.baseItemLabelGenerator = new StandardCategoryItemLabelGenerator(StandardCategoryItemLabelGenerator.DEFAULT_LABEL_FORMAT_STRING, hourFormat, hourFormat)
			chart.plot.renderer.baseItemLabelsVisible = true

			chart.plot.rangeAxis.upperMargin = 0.1D
			chart.plot.rangeAxis.visible = false

			chart.plot.outlinePaint = null
		}
	}

	PieDataset getTotalTimeSpentPieDataset(Map totalTimeSpentMetrics) {
		def dataset = new DefaultPieDataset()
		for (metrics in totalTimeSpentMetrics.entrySet()) {
			dataset.setValue(metrics.key, metrics.value["percentage"] * 100)
		}

		return dataset
	}

	def getTotalTimeSpentPieChartCustomizer() {
		return { chart ->
			fillPieChartColors(chart)

			chart.plot.outlinePaint = null

			chart.plot.labelGenerator = new StandardPieSectionLabelGenerator("{0}\n({2})", NumberFormat.getNumberInstance(), percentageFormat)
		}
	}

	private Map getDurations(def issue) {
		def start = issue.getCreated()
		def durations = new LinkedHashMap()
		def lastValue

		def changeHistories = componentAccessor.changeHistoryManager.getChangeHistories(issue)
		for (changeHistory in changeHistories) {
			for (changeItem in changeHistory.changeItemBeans) {
				if (isRelatedChangeItem(changeItem)) {
					def key = normalizeValue(changeItem.fromString)
					durations.putIfAbsent(key, [])
					durations.get(key).add(getDurationInMillis(start, changeHistory.timePerformed))

					lastValue = changeItem.toString
					start = changeHistory.timePerformed
				}
			}
		}

		lastValue = lastValue ?: getCurrentFieldValue(issue)
		durations.putIfAbsent(lastValue, [])
		durations.get(lastValue).add(getDurationInMillis(start, new Date()))

		return durations
	}

	private getIssueFirstResponseTime(issue) {
		def comments = componentAccessor.commentManager.getCommentsForUser(issue, user)

		return !comments.empty ? comments[0].created : null
	}

	private static def getIssueResolutionTime(issue) {
		issue.resolutionDate // returns NULL for unresolved
	}

	private static def getDurationInMillis(Date from, Date to) {
		to.getTime() - from.getTime()
	}

	private def getDuration(Date from, Date to) {
		def millis = getDurationInMillis(from, to)
		def hours = millis / 1000 / 60 / 60

		if (!prettyPrint) {
			return hourFormat.format(hours)
		}

		def duration = TimeCategory.minus(to, from)
		return hourFormatForPrettyPrint.format(hours) + " " + duration.minutes + "m " + duration.seconds + "s"
	}

	private static getAxisTickLabelFormat(long startMillis, long endMillis, boolean showHours) {
		final long hoursSpanned = getHoursSpanned(startMillis, endMillis)
		String formatString
		boolean isLong = false
		if (hoursSpanned < ONE_DAY_IN_HOURS) {
			formatString = "H:mm"
		} else if (hoursSpanned < ONE_YEAR_IN_HOURS) {
			formatString = "MMM dd" + (showHours ? " H:mm" : "")
			isLong = showHours
		} else {
			formatString = "YYYY MMM dd" + (showHours ? " H:mm" : "")
			isLong = true
		}
		return [dateFormat: new SimpleDateFormat(formatString), isLong: isLong]
	}

	private static boolean isAtRoundDay(double value) {
		(value % DateTimeConstants.SECONDS_PER_DAY) == 0
	}

	private static int getHoursSpanned(long startMillis, long endMillis) {
		Hours.hoursBetween(new DateTime(startMillis), new DateTime(endMillis)).getHours()
	}
}